package abc.ltl;

import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import polyglot.ast.NodeFactory;
import polyglot.frontend.Compiler;
import polyglot.frontend.CupParser;
import polyglot.frontend.FileSource;
import polyglot.frontend.Job;
import polyglot.frontend.Parser;
import polyglot.frontend.Pass;
import polyglot.frontend.PrettyPrintPass;
import polyglot.frontend.VisitorPass;
import polyglot.lex.Lexer;
import polyglot.util.CodeWriter;
import polyglot.util.ErrorQueue;
import polyglot.visit.PrettyPrinter;
import abc.aspectj.ast.AJNodeFactory;
import abc.aspectj.parse.Lexer_c;
import abc.aspectj.visit.NoSourceJob;
import abc.ltl.ast.LTLNodeFactory_c;
import abc.ltl.formulaextraction.FormulaSource;
import abc.ltl.formulaextraction.Specification;
import abc.ltl.parse.Grm;
import abc.ltl.visit.override.CheckPackageNames;
import abc.ltl.visit.override.CollectJimplifyVisitor;
import abc.ltl.visit.transform.AdviceGeneration;
import abc.ltl.visit.transform.CleanUp;
import abc.ltl.visit.transform.ConstructorDeclaration;
import abc.ltl.visit.transform.GeneralFieldDecls;
import abc.ltl.visit.transform.ImportStatements;
import abc.ltl.visit.transform.MultiThreadingSecure;
import abc.ltl.visit.transform.PropositionLabels;
import abc.main.CompilerFailedException;

/**
 * Extension information for LTL extension.
 * @author Eric Bodden
 */
public class ExtensionInfo extends abc.meta.ExtensionInfo
{

    public static final polyglot.frontend.Pass.ID CONVERT_NEGATIONS = new polyglot.frontend.Pass.ID("convert-negations");
    public static final polyglot.frontend.Pass.ID IMPORTS = new polyglot.frontend.Pass.ID("imports");
    public static final polyglot.frontend.Pass.ID GENERAL_FIELD_DECLS = new polyglot.frontend.Pass.ID("general-field-declarations");
    public static final polyglot.frontend.Pass.ID PROPOSITION_LABELS = new polyglot.frontend.Pass.ID("proposition-labels");
    public static final polyglot.frontend.Pass.ID CONSTRUCTOR_DECL = new polyglot.frontend.Pass.ID("constructor-decl");
    public static final polyglot.frontend.Pass.ID PROPOSITION_ADVICE = new polyglot.frontend.Pass.ID("proposition-advice");
    public static final polyglot.frontend.Pass.ID CLEANUP = new polyglot.frontend.Pass.ID("cleanup");
	public static final polyglot.frontend.Pass.ID FORMULA_CONSTRUCTORS = new polyglot.frontend.Pass.ID("formula-constructors");
    public static final polyglot.frontend.Pass.ID FORMULA_IDS = new polyglot.frontend.Pass.ID("formula-ids");
    public static final polyglot.frontend.Pass.ID MULTI_THREADING_SECURE = new polyglot.frontend.Pass.ID("multi-threading-secure");
    public static final polyglot.frontend.Pass.ID LOAD_FORMULA_SOOT_CLASSES = new polyglot.frontend.Pass.ID("load-formula-soot-classes");

    static {
        // force Topics to load
        new Topics();
    }

    public ExtensionInfo(Collection jar_classes, Collection source_files)
    {
        super(jar_classes, source_files);
    }

    /** 
	 * @inheritDoc
	 */
    @Override
	public void initCompiler(Compiler comp) {
		super.initCompiler(comp);
        //load formulas into specification and add appropriate source units to
        //polyglot's jobs
        Specification.generateInitialSourceUnitsFromFormulas(this);
	}

    @Override
    public String compilerName() {
        return "ltl";
    }

    @Override
    public Parser parser(Reader reader, FileSource source, ErrorQueue eq) {
        Lexer lexer = new Lexer_c(reader, source.name(), eq);
        Grm grm = new Grm(lexer, ts, nf, eq);
        return new CupParser(grm, source, eq);
    }

    @Override
    protected NodeFactory createNodeFactory() {
        return new LTLNodeFactory_c();
    }
       
    /**
     * {@inheritDoc}
     */
    @Override
    public List passes(Job job) {

        //workaround: when no source files are given on the commandline, abc usually passes in a 
        //NoSourceJob to just execute all RunOnce passes. However, when we have to execute the full
        //compilation process, when formulae exist  
        if(job instanceof NoSourceJob && Specification.getFormulas().size()>0) {
            if (!compiler.compile(Collections.EMPTY_LIST)) {
                throw new RuntimeException(new CompilerFailedException("Compilation failed."));
            }
            return Collections.EMPTY_LIST;
        }
        
        //first delegate up, then modify the list
        List l = super.passes(job);
        
        List<Pass> afterParsing = new ArrayList<Pass>();
        
        if(!(job instanceof NoSourceJob) &&
                job.source() instanceof FormulaSource) {
            //if we have a formula source file, transform it

            afterParsing.add(new VisitorPass(IMPORTS, job, new ImportStatements(nf)));
            afterParsing.add(new VisitorPass(GENERAL_FIELD_DECLS, job, new GeneralFieldDecls(nf)));
            afterParsing.add(new VisitorPass(PROPOSITION_LABELS, job, new PropositionLabels((AJNodeFactory)nf,ts)));
            afterParsing.add(new VisitorPass(CONSTRUCTOR_DECL, job, new ConstructorDeclaration(nf)));
            
            if(!(nf instanceof AJNodeFactory)) {
                throw new RuntimeException("Unexpected type of node factory (need AJNodeFactory): " + nf.getClass().getCanonicalName());
            }            
            afterParsing.add(new VisitorPass(PROPOSITION_ADVICE, job, new AdviceGeneration((AJNodeFactory)nf,ts)));
            afterParsing.add(new VisitorPass(MULTI_THREADING_SECURE, job, new MultiThreadingSecure(nf)));
            afterParsing.add(new VisitorPass(CLEANUP, job, new CleanUp()));
            
            if(Debug.v().showGeneratedAspects) {
                afterParsing.add(new PrettyPrintPass(INSPECT_AST,job,new CodeWriter(System.out,1000),new PrettyPrinter()));
            }

            //insert the passes
            afterPass(l, Pass.PARSE, afterParsing);
        }
        
        //in any case, override the tow necessary visitors...

        //first the CollectJimplifyVisitor
        List<Pass> newCollectJimplify = new ArrayList<Pass>();
        newCollectJimplify.add(new VisitorPass(COLLECT_JIMPLIFY_CLASSES, job,
                new CollectJimplifyVisitor(job, ts, nf, source_files, class_to_ast)));        
        replacePass(l, COLLECT_JIMPLIFY_CLASSES, newCollectJimplify);

        //then CheckPackageNames
        List<Pass> newCheckPackageNames = new ArrayList<Pass>();
        newCheckPackageNames.add(new CheckPackageNames(CHECK_PACKAGE_NAMES,job));
        replacePass(l, CHECK_PACKAGE_NAMES, newCheckPackageNames);

        return l;
    }
	
}
