package abc.ltl.ast;

import static abc.ltl.ast.EntryExit.Exit.ExitKind.RETURNING;
import static abc.ltl.ast.EntryExit.Exit.ExitKind.THROWING;
import static abc.ltl.visit.transform.RuntimeRepresentation.BINDINGS_MAP_DECLARED_TYPE_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.BINDINGS_MAP_LOCAL_VARIABLE_GET_METHOD;
import static abc.ltl.visit.transform.RuntimeRepresentation.FACTORY_LABEL;
import static abc.ltl.visit.transform.RuntimeRepresentation.PROPOSITION_CLASS_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.PROPOSITION_IF_CLOSURE_CLASS_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.PROPOSITION_IF_CLOSURE_EVAL_METHOD_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.PROPOSITION_IF_CLOSURE_VARIABLES_METHOD_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.USER_CAUSED_EXCEPTION_CLASS_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.USER_CAUSED_EXCEPTION_QUALIFIER_NAME;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import polyglot.ast.AmbExpr;
import polyglot.ast.AmbReceiver;
import polyglot.ast.AmbTypeNode;
import polyglot.ast.Block;
import polyglot.ast.CanonicalTypeNode;
import polyglot.ast.Catch;
import polyglot.ast.ClassMember;
import polyglot.ast.Expr;
import polyglot.ast.Formal;
import polyglot.ast.Node;
import polyglot.ast.NodeFactory;
import polyglot.ast.Stmt;
import polyglot.ast.TypeNode;
import polyglot.types.Flags;
import polyglot.types.PrimitiveType;
import polyglot.types.TypeSystem;
import polyglot.util.Position;
import polyglot.util.TypedList;
import polyglot.visit.NodeVisitor;
import abc.aspectj.ast.AJNodeFactory;
import abc.aspectj.ast.AmbTypeOrLocal;
import abc.aspectj.ast.PCIf;
import abc.aspectj.ast.Pointcut;
import abc.ltl.Util;
import abc.ltl.ast.override.StatebindingPointcut;
import abc.ltl.visit.transform.PropositionLabels;

/**
 * PointcutUtils
 * 
 * @author Eric Bodden
 */
public class PointcutUtils {

    /**
     * @param pc
     *            pointcut to generate the constructor expression for
     * @param vis
     *            reference to visitor
     * @param pos
     *            position information for the generated expression
     * @param declaredFormals
     *            formals declared in the enclosing formula
     * @return the instantiation expression
     */
    public static Expr propositionInstantiationExprForPointcut(Pointcut pc,
            PropositionLabels vis, Collection<Formal> declaredFormals,
            Position pos) {
        // should only be used for *simple* pointcuts
        assert !(pc instanceof PCLTLGeneral);

        AJNodeFactory nf = vis.nf();
        TypeSystem ts = vis.ts();

        // first argument: the texttual representation of the pointcut
        Expr pointcutId = nf.StringLit(pos, EntryExit.pcToString(pc));

        // second argument: formals passed in
        List<Expr> formalsArgs = new ArrayList<Expr>();

        // add formals
        for (Formal formal : boundFormals(declaredFormals,pc,false)) {
            formalsArgs.add(nf.StringLit(pos, formal.name()));
        }

        //handle type/formal in returning/throwing statement
        if(EntryExit.isExit(pc)) {
            EntryExit.Exit exit = (EntryExit.Exit) EntryExit.getEntryExit(pc);
            if(exit.getKind().equals(RETURNING) || exit.getKind().equals(THROWING)) {
                //if type name is a declared formal, add it to the list
                for (Formal formal : declaredFormals) {
                    if(formal.name().equals(exit.getTypeName())) {
                        formalsArgs.add(nf.StringLit(pos, formal.name()));
                    }
                }
            }
        }

        Expr ifClosuresArrayArg = exprForIfClosures(pc, declaredFormals, pos, nf, ts);        

        //create array for "formal" labels
        Expr formalArrayArg = nf.NewArray(
                pos,
                nf.AmbTypeNode(
                        pos,
                        "java.lang.String"
                ),                        
                1,
                nf.ArrayInit(
                        pos,
                        formalsArgs
                )
        );        
        
        return nf.Call(
                pos,
                nf.AmbReceiver(
                        pos,
                        FACTORY_LABEL
                ),
                PROPOSITION_CLASS_NAME,
                pointcutId,
                formalArrayArg,
                ifClosuresArrayArg
        );
    }

    /**
     * @param pc
     * @param declaredFormals
     * @param pos
     * @param nf
     * @param ts
     * @return
     */
    @SuppressWarnings("unchecked")
    private static Expr exprForIfClosures(Pointcut pc, Collection<Formal> declaredFormals, Position pos, AJNodeFactory nf, TypeSystem ts) {
        // third argument: if-closures passed in 
        List<Expr> ifClosuresArgs = new ArrayList<Expr>();
        
        Collection<IfPointcutCollector.ExpressionPlusBindings> expressionsPlusBindings =
            collectIfPointcuts(pc, declaredFormals, false, nf);
        
        //only for internal use here
        final String BINDINGS_MAP_NAME_IN_IF_CLOSURE = "currentBindings";
        
        for (IfPointcutCollector.ExpressionPlusBindings exprPlusBinding : expressionsPlusBindings) {
            Expr expr = exprPlusBinding.getExpr();            
            Collection<Formal> boundFormals = exprPlusBinding.getBoundFormals();
            
            List formals = new ArrayList();
            formals.add(
                    nf.Formal(
                            pos,
                            Flags.NONE,
                            nf.AmbTypeNode(
                                    pos,
                                    BINDINGS_MAP_DECLARED_TYPE_NAME
                            ),
                            BINDINGS_MAP_NAME_IN_IF_CLOSURE
                    )
            );
            //seal list
            formals = new TypedList(formals,Formal.class,true);            
            
            List statements = new ArrayList();
            for (Formal formal : boundFormals) {
                Expr init = nf.Call(
                        pos,
                        nf.AmbReceiver(
                                pos,
                                BINDINGS_MAP_NAME_IN_IF_CLOSURE
                        ),
                        BINDINGS_MAP_LOCAL_VARIABLE_GET_METHOD,
                        nf.StringLit(
                                pos,
                                formal.name()
                        )
                );
                
                //perform casts and autounboxing
                init = autoUnBox(init, formal.type(), nf);
                
                statements.add(
                        nf.LocalDecl(
                                pos,
                                Flags.FINAL,
                                formal.type(),
                                formal.name(),
                                init
                        )
                );
            }
            
            //internally used only
            final String USER_CAUSED_EXCEPTION_LABEL = "ex";
            
            //arguments for the UserCausedException
            List userCausedExceptionArgs = new ArrayList();
            //first argument: the exception
            userCausedExceptionArgs.add(
                    nf.AmbExpr(pos,USER_CAUSED_EXCEPTION_LABEL)
            );
            //second argument: the expression as string representation
            userCausedExceptionArgs.add(
                    nf.StringLit(pos,Util.nodeToString(expr))
            );
            //seal list
            userCausedExceptionArgs = new TypedList(userCausedExceptionArgs,Expr.class,true);
            
            //inner catch block
            List innerCatchBlocks = new ArrayList();
            //catch(Exception ex) { throw new UserCausedException(<expr>,ex); }
            innerCatchBlocks.add(
                    nf.Catch(
                            pos,
                            nf.Formal(
                                    pos,
                                    Flags.NONE,
                                    nf.AmbTypeNode(
                                            pos,
                                            "java.lang.Exception"
                                    ),
                                    USER_CAUSED_EXCEPTION_LABEL
                            ),
                            nf.Block(
                                    pos,
                                    nf.Throw(
                                            pos,
                                            nf.New(
                                                    pos,
                                                    nf.AmbTypeNode(
                                                            pos,
                                                            nf.AmbQualifierNode(
                                                                    pos,
                                                                    USER_CAUSED_EXCEPTION_QUALIFIER_NAME
                                                            ),
                                                            USER_CAUSED_EXCEPTION_CLASS_NAME
                                                    ),
                                                    userCausedExceptionArgs
                                            )
                                    )
                            )
                    )
            
            );
            //seal list
            innerCatchBlocks = new TypedList(innerCatchBlocks,Catch.class,true);
            
            //try { return <expr>; } ...
            statements.add(
                    nf.Try(
                            pos,
                            nf.Block(
                                    pos,
                                    nf.Return(
                                            pos,
                                            expr
                                    )
                            ),
                            innerCatchBlocks                            
                    )
            );
            //seal list
            statements = new TypedList(statements,Stmt.class,true);
            
            Block blockInsideOuterTry = nf.Block(
                    pos,
                    statements
            );

            List outerCatchBlocks = new ArrayList();
            
            //add try/catch blocks for the following exceptions
            //which all just return "false"
            for(String exceptionName: new String[] {
                    "java.lang.NullPointerException",
                    "java.lang.ClassCastException"
            }) {
                outerCatchBlocks.add(
                        nf.Catch(
                                pos,
                                nf.Formal(
                                        pos,
                                        Flags.NONE,
                                        nf.AmbTypeNode(
                                                pos,
                                                exceptionName
                                        ),
                                        "ex"
                                ),
                                nf.Block(
                                        pos,
                                        nf.Return(
                                                pos,
                                                nf.BooleanLit(
                                                        pos,
                                                        false
                                                )
                                        )
                                )
                        )
                );
            }
            blockInsideOuterTry = nf.Block(
                    pos,
                    nf.Try(
                            pos,
                            blockInsideOuterTry,
                            outerCatchBlocks
                    )
            );
            
            List members = new ArrayList();
            //add evaluation method
            members.add(
                    nf.MethodDecl(
                            pos,
                            Flags.PUBLIC,
                            nf.CanonicalTypeNode(
                                    pos,
                                    ts.Boolean()
                            ),
                            PROPOSITION_IF_CLOSURE_EVAL_METHOD_NAME,
                            formals,
                            Collections.EMPTY_LIST,
                            blockInsideOuterTry
                    )
            );
            
            //variable names bound by one of the if-closures AND by the pointcut itself 
            List arrayArgs = new ArrayList();
            for (Formal formal : boundFormals) {            	
            	arrayArgs.add(
						nf.StringLit(
								pos,
								formal.name()
						)
				);
			}
            //seal list
            arrayArgs = new TypedList(arrayArgs,Expr.class,true);
            
			//add method returning the variable names for this if-pointcut 
            members.add(
                    nf.MethodDecl(
                            pos,
                            Flags.PUBLIC,
                            nf.ArrayTypeNode(
                            	pos,
                            	nf.AmbTypeNode(
                            			pos,
                            			"java.lang.String"
                            	)
                            ),
                            PROPOSITION_IF_CLOSURE_VARIABLES_METHOD_NAME,
                            Collections.EMPTY_LIST,
                            Collections.EMPTY_LIST,
                            nf.Block(
                            		pos,
                            		nf.Return(
                            				pos,
                            				nf.NewArray(
                            						pos,
                                                	nf.AmbTypeNode(
                                                			pos,
                                                			"java.lang.String"
                                                	),
                                                	1,
                                                	nf.ArrayInit(
                                                			pos,
                                                			arrayArgs
                                                	)
                            				)
                            		)
                            )
                    )
            );
            //add toString method (for debugging)
            members.add(
                    nf.MethodDecl(
                            pos,
                            Flags.PUBLIC,
                            nf.AmbTypeNode(
                                    pos,
                                    "java.lang.String"
                            ),
                            "toString",
                            Collections.EMPTY_LIST,
                            Collections.EMPTY_LIST,
                            nf.Block(
                                    pos,
                                    nf.Return(
                                            pos,
                                            nf.StringLit(
                                                    pos,
                                                    Util.nodeToString(expr)
                                            )
                                    )
                            )
                    )
            );
            //seal list
            members = new TypedList(members,ClassMember.class,true);            
            
            ifClosuresArgs.add(
                    nf.New(
                            pos,
                            nf.AmbTypeNode(
                                    pos,
                                    PROPOSITION_IF_CLOSURE_CLASS_NAME
                            ),                        
                            Collections.EMPTY_LIST,
                            nf.ClassBody(
                                    pos,
                                    members
                            )
                    )
            );
        }
        
        //create array for if-closures
        Expr ifClosuresArrayArg = nf.NewArray(
                pos,
                nf.AmbTypeNode(
                        pos,
                        PROPOSITION_IF_CLOSURE_CLASS_NAME
                ),                        
                1,
                nf.ArrayInit(
                        pos,
                        ifClosuresArgs
                )
        );
        return ifClosuresArrayArg;
    }

    /**
     * @param declaredFormals
     *            formals declared in the formula
     * @param pc
     *            pointcut
     * @param includeFormalsBoundByIfPCs
     * 			  if <code>true</code>, formals bound by if-subpointcuts are included
     * @return subset of <i>declaredFormals</i>, holding all formals bound by
     *         the pointcut <i>pc</i>
     */
    public static List<Formal> boundFormals(Collection<Formal> declaredFormals,
            Pointcut pc, boolean includeFormalsBoundByIfPCs) {
        List<Formal> boundFormals = new ArrayList<Formal>();

        //collect all bindings
        BindingCollector bc = new BindingCollector();
        pc.visit(bc);
        
        //add all formals which match a bound name
        for (Formal formal : declaredFormals) {
            if(bc.boundNames().contains(formal.name())) {
                boundFormals.add(formal);
            }
        }
        
        if(includeFormalsBoundByIfPCs) {
	        IfPointcutCollector ipc = new IfPointcutCollector(declaredFormals);
	        pc.visit(ipc);
	        
	        for (IfPointcutCollector.ExpressionPlusBindings exprPlusBinding :
	        	ipc.getExpressionsPlusBindings()) {
	        	for (Formal formal : exprPlusBinding.getBoundFormals()) {
					if(!boundFormals.contains(formal)) {
						boundFormals.add(formal);
					}
				}
			}
        }
        
        return boundFormals;
    }
    
    public static Pointcut replaceBindingsByTypes(Collection<Formal> declaredFormals,
            Pointcut pc) {
        //collect all bindings
        BindingReplacer bc = new BindingReplacer(declaredFormals);
        return (Pointcut) pc.visit(bc);
    }
    
    /**
     * Helper class to collect bindings from this/target/args.
     * 
     * @author Eric Bodden
     */
    private static class BindingCollector extends NodeVisitor {

        private List<String> boundNames  = new ArrayList<String>();
        
        /**
         * {@inheritDoc}
         */
        @Override
        public NodeVisitor begin() {
            this.boundNames = new ArrayList<String>();
            return super.begin();
        }
        
        /**
         * {@inheritDoc}
         */
        @Override
        public NodeVisitor enter(Node n) {
            if(n instanceof StatebindingPointcut) {
                Node[] patterns = ((StatebindingPointcut)n).patterns();
                
                for (Node node : patterns) {
                    if(!(node instanceof AmbTypeOrLocal)) {
                        throw new RuntimeException("internal error in BindingCollector [1]");
                    }
                    TypeNode type = ((AmbTypeOrLocal)node).type();
                    
                    if(!(type instanceof AmbTypeNode)) {
                        throw new RuntimeException("internal error in BindingCollector [2]");
                    }
                    AmbTypeNode amb = (AmbTypeNode) type;
                    
                    this.boundNames.add(amb.name());                    
                }
                
            }
            return super.enter(n);
        }
        
        public List<String> boundNames() {
            return Collections.unmodifiableList(this.boundNames);
        }
        
    }

    /**
     * Helper class to collect bindings from this/target/args.
     * 
     * @author Eric Bodden
     */
    private static class BindingReplacer extends NodeVisitor {
        
        private Collection<Formal> declaredFormals;
        
        /**
         * @param declaredFormals
         */
        public BindingReplacer(Collection<Formal> declaredFormals) {
            this.declaredFormals = declaredFormals;
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public Node leave(Node old, Node n, NodeVisitor v) {
            if(n instanceof StatebindingPointcut) {
                n = ((StatebindingPointcut)n).replaceBindingsByTypes((Formal[])this.declaredFormals.toArray(new Formal[0]));
            }
            return super.leave(old, n, v);
        }
                
    }
    
    public static TypeNode bindingNameToType(String bindingName, Formal... declaredFormals) {
        for (Formal formal : declaredFormals) {
            if(formal.name().equals(bindingName)) {
                return formal.type();
            }
        }
        return null;
    }

    private static Collection<Formal> formalsBoundByIfPointcut(PCIf pc, Collection<Formal> declaredFormals) {
        IfPointcutBindingCollector vs = new IfPointcutBindingCollector(declaredFormals);
        pc.visit(vs);
        return vs.getBoundFormals();
    }
    
    private static class IfPointcutBindingCollector extends NodeVisitor {

        private Collection<Formal> declaredFormals;
        
        private Collection<Formal> boundFormals;
        
        /**
         * @param pc
         * @param declaredFormals
         */
        public IfPointcutBindingCollector(Collection<Formal> declaredFormals) {
            this.declaredFormals = declaredFormals;
            this.boundFormals = new ArrayList<Formal>();
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public Node leave(Node old, Node n, NodeVisitor v) {            
        	if(n instanceof AmbExpr) {
                AmbExpr amb = (AmbExpr) n;
                for (Formal formal : declaredFormals) {
                    if(formal.name().equals(amb.name())) {
                        boundFormals.add(formal);
                    }
                }
            } else if(n instanceof AmbReceiver) {
                AmbReceiver amb = (AmbReceiver) n;
                for (Formal formal : declaredFormals) {
                    if(formal.name().equals(amb.name())) {
                        boundFormals.add(formal);
                    }
                }
            }
            return super.leave(old, n, v);
        }

        public Collection<Formal> getBoundFormals() {
            return boundFormals;
        }
        
    }
    
    private static Collection<IfPointcutCollector.ExpressionPlusBindings> collectIfPointcuts(
            Pointcut pc, Collection<Formal> declaredFormals, boolean replaceByIfTrue, AJNodeFactory nf)
    {
        IfPointcutCollector vs = new IfPointcutCollector(declaredFormals, replaceByIfTrue, nf);
        pc.visit(vs);
        return vs.getExpressionsPlusBindings();
    }
    
    public static Pointcut replaceIfPCsByIfTrue(Pointcut pc, AJNodeFactory nf) {
        final Collection<Formal> emptyList = Collections.emptyList();
        IfPointcutCollector vs = new IfPointcutCollector(emptyList, true, nf);
        return (Pointcut)pc.visit(vs);
    }
    
    private static class IfPointcutCollector extends NodeVisitor {
        
        private static class ExpressionPlusBindings {
            
            private Expr expr;
            
            private Collection<Formal> boundFormals;

            /**
             * @param expr
             * @param formals
             */
            private ExpressionPlusBindings(Expr expr, Collection<Formal> formals) {
                this.expr = expr;
                boundFormals = formals;
            }

            public Collection<Formal> getBoundFormals() {
                return boundFormals;
            }

            public Expr getExpr() {
                return expr;
            }
            
        }
        
        private boolean replaceByIfTrue;
        
        private AJNodeFactory nf;
        
        private Collection<Formal> declaredFormals;
        
        private Collection<ExpressionPlusBindings> result;

        /**
         * @param declaredFormals
         * @param replaceByIfTrue
         * @param nf
         */
        private IfPointcutCollector(Collection<Formal> declaredFormals, boolean replaceByIfTrue, AJNodeFactory nf) {
            this. replaceByIfTrue = replaceByIfTrue;
            this.nf = nf;
            this.declaredFormals = declaredFormals;
            this.result = new ArrayList<ExpressionPlusBindings>();
        }
        
        /**
         * @param declaredFormals
         */
        private IfPointcutCollector(Collection<Formal> declaredFormals) {
            this(declaredFormals,false,null);
        }

        /**
         * {@inheritDoc}
         */
        @Override
        public Node visitEdge(Node parent, Node child) {
            if(parent instanceof PCIf) {
                if(child instanceof Expr) {
                    PCIf pc = (PCIf) parent;
                    Expr expr = (Expr) child;
                    Collection<Formal> boundFormals = formalsBoundByIfPointcut(pc,declaredFormals);
                    result.add(new ExpressionPlusBindings(expr,boundFormals)); 
                }
                return child;
            }
            return super.visitEdge(parent,child);
        }
        
        /**
         * {@inheritDoc}
         */
        @Override
        public Node leave(Node old, Node n, NodeVisitor v) {
            if(n instanceof PCIf) {
                if(replaceByIfTrue) {
                    //replace if-pointcut by an "if(true)" pointcut
                    n = nf.PCIf(n.position(),nf.BooleanLit(n.position(),true));
                }
            }
            return super.leave(old, n, v);
        }

        public Collection<ExpressionPlusBindings> getExpressionsPlusBindings() {
            return result;
        }
        
        
    }
    
    private static Expr autoUnBox(Expr expr, TypeNode targetType, NodeFactory nf) {
        
        Position pos = expr.position();
        
        String unboxMethodName = null;
        TypeNode castType;
        
        //do we have to cast to a boxing type ?
        if(targetType instanceof CanonicalTypeNode) {
            CanonicalTypeNode tn = (CanonicalTypeNode) targetType;
            if(tn.type() instanceof PrimitiveType) {
                PrimitiveType pt = (PrimitiveType) tn.type();
                String boxTypeName;
                if(pt.kind().equals(PrimitiveType.BOOLEAN)) {
                    boxTypeName = "Boolean";
                    unboxMethodName = "booleanValue";
                } else if(pt.kind().equals(PrimitiveType.BYTE)) {
                    boxTypeName = "Byte";
                    unboxMethodName = "byteValue";
                } else if(pt.kind().equals(PrimitiveType.CHAR)) {
                    boxTypeName = "Character";
                    unboxMethodName = "charValue";
                } else if(pt.kind().equals(PrimitiveType.DOUBLE)) {
                    boxTypeName = "Double";
                    unboxMethodName = "doubleValue";
                } else if(pt.kind().equals(PrimitiveType.FLOAT)) {
                    boxTypeName = "Float";
                    unboxMethodName = "floatValue";
                } else if(pt.kind().equals(PrimitiveType.INT)) {
                    boxTypeName = "Integer";
                    unboxMethodName = "intValue";
                } else if(pt.kind().equals(PrimitiveType.LONG)) {
                    boxTypeName = "Long";
                    unboxMethodName = "longValue";
                } else if(pt.kind().equals(PrimitiveType.SHORT)) {
                    boxTypeName = "Short";
                    unboxMethodName = "shortValue";
                } else {
                    throw new RuntimeException("Unexpected primitive type in 'PointcutUtils' (autounboxing)");
                }
                castType = nf.AmbTypeNode(pos,"java.lang." + boxTypeName);
            } else {
                throw new RuntimeException("Unexpected canonical type in 'PointcutUtils' (autounboxing)");
            }
        } else {
            castType = targetType;
        }
         
        //cast the expression
        expr = nf.Cast(
                pos,
                castType,
                expr
        );

        //if we have a primitive type, we have to get the primitive value (unboxing)
        if(unboxMethodName != null) {
        
            expr = nf.Call(
                    pos,
                    expr,
                    unboxMethodName                    
            );

        }
        
        return expr;
        
    }
    
    
}
