package abc.ltl.formulaextraction;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * @author Eric Bodden
 *
 * AnnotationExtractor
 * 
 * SAX handler for searching for annotations in the XML output created by BAT2XML.
 * 
 */
/* package */ class AnnotationExtractor extends DefaultHandler {
    
    private enum State {INITIAL, INSIDE_ANNO, INSIDE_LTL_ANNO, INSIDE_LTL_VALUE, INSIDE_SIGNATURE, CODE};
    
    private boolean inInherits;
    
    private State state;
    
    private String classId;
    
    private String sourceFile;

    private String annotatedItemName;
    
    private String annotation;
    
    private int lastLineNumber;
    
    /**
     * {@inheritDoc}
     */
    public void startDocument() throws SAXException {
        state = State.INITIAL;
        classId = null;
        sourceFile = null;
        annotatedItemName = null;
        inInherits = false;
    }
    
     /**
     * {@inheritDoc}
     */
    public void startElement(
            @SuppressWarnings("unused") String uri,
            @SuppressWarnings("unused") String localName,
            String qName,
            Attributes attributes) throws SAXException {
        if(qName.equals("code")) {
            state = State.CODE;
        }
        if((state == State.INITIAL || state == State.INSIDE_ANNO || state == State.INSIDE_LTL_ANNO) && qName.equals("class")
        && !this.inInherits) {
            this.annotation = null;
            this.lastLineNumber = 1;
            int index = attributes.getIndex("name");
            assert attributes.getValue(index) != null;
            this.classId = attributes.getValue(index);
            index = attributes.getIndex("sourcefile");
            if(index>-1) {
                this.sourceFile = attributes.getValue(index);
            } else {
                this.sourceFile = "";
            }
            annotatedItemName = this.classId; 
        } else if((state == State.INITIAL) &&
                ( qName.equals("method") || qName.equals("field") ) 
                ) {
            this.annotation = null;
            int index = attributes.getIndex("name");
            assert attributes.getValue(index) != null;
            String name = attributes.getValue(index);
            //fully qualify the method/constructor/field
            this.annotatedItemName = this.classId + "." + name;
            if(qName.equals("method")) {
                String prefix = attributes.getValue(attributes.getIndex("visibility"));
                if(prefix.length()>0)
                    prefix += " ";
                if(attributes.getValue(attributes.getIndex("abstract")).equals("true"))
                    prefix += "abstract ";
                if(attributes.getValue(attributes.getIndex("static")).equals("true"))
                    prefix += "static ";
                if(attributes.getValue(attributes.getIndex("synchronized")).equals("true"))
                    prefix += "synchronized ";
                this.annotatedItemName = prefix + annotatedItemName + "(";
            } else {
                //field
                this.annotatedItemName = attributes.getValue(attributes.getIndex("type")) +
                    " " + this.annotatedItemName; 
                
            }
        } else if(state == State.INITIAL && qName.equals("annotation")) {
            if(attributes.getIndex("xmlns")>-1) {
                //annotation *types* are also denoted by "annotation"
                //skip those
                return;
            }
            state = State.INSIDE_ANNO;
            //found annotation
            int index = attributes.getIndex("type");
            if(index > -1) {
                if(attributes.getValue(index).equals(Specification.LTL_TYPE_NAME)) {
                    state = State.INSIDE_LTL_ANNO;
                }
            }
        } else if((state == State.INITIAL) && qName.equals("signature")) {
            state = State.INSIDE_SIGNATURE;
        } else if((state == State.INSIDE_SIGNATURE) && qName.equals("returns") &&
                !this.annotatedItemName.contains("<init>")) {
            String returnType = attributes.getValue(attributes.getIndex("type"));
            String methodName = this.annotatedItemName.substring(this.annotatedItemName.lastIndexOf(" ")+1);
            String prefix = this.annotatedItemName.substring(0,this.annotatedItemName.lastIndexOf(" ")+1);
            this.annotatedItemName = prefix + returnType + " " + methodName;
        } else if((state == State.INSIDE_SIGNATURE) && qName.equals("parameter")) {
            String type = attributes.getValue(attributes.getIndex("type"));
            if(this.annotatedItemName.charAt(this.annotatedItemName.length()-1)!='(') {
                this.annotatedItemName += ",";
            }
            this.annotatedItemName += type;
        } else if(state == State.INSIDE_LTL_ANNO && qName.equals("value")) {
            this.annotation = "";
            state = State.INSIDE_LTL_VALUE;
        } else if(qName.equals("inherits")) {
            //avoid parsing of super class IDs
            this.inInherits = true;
        }
        if(state == State.CODE && attributes.getIndex("lineNumber")>-1) {
            this.lastLineNumber = Integer.parseInt(attributes.getValue("lineNumber"));
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public void characters(char[] ch, int start, int length)
            throws SAXException {
        if(state == State.INSIDE_LTL_VALUE) {
            this.annotation += new String(ch, start, length);
        }
    }
    
    /**
     * {@inheritDoc}
     */
    public void endElement(
            @SuppressWarnings("unused") String uri,
            @SuppressWarnings("unused") String localName,
            String qName)
            throws SAXException {
        if(state == State.INSIDE_LTL_VALUE) {
            //leaving value
            state = State.INSIDE_LTL_ANNO;
        } else if((state == State.INSIDE_LTL_ANNO || state == State.INSIDE_ANNO)
                && qName.equals("annotation")) {
            //leaving annotation
            state = State.INITIAL;
        } else if(state == State.INSIDE_SIGNATURE && qName.equals("signature")) {
            if(this.annotatedItemName.contains("(")) {
                this.annotatedItemName += ")";
            }
            //leaving signature
            state = State.INITIAL;
        } else if(state == State.CODE && qName.equals("code")) {
            state = State.INITIAL;
        }
        if(qName.equals("method") || qName.equals("field") || qName.equals("class")) {
            if(this.annotation!=null && this.annotatedItemName!=null) {
                String formula = this.annotation.replaceAll("\\$", ".");
                String insert = this.annotatedItemName.replaceAll("\\$", ".");
                formula = formula.replaceAll("thisType",this.classId.replaceAll("\\$", "."));
                formula = formula.replaceAll("thisMember",insert);
                formula = formula.replaceAll("<init>","new");
                
                String typeName = this.classId.replaceAll("\\$", ".");
                Specification.addFormula(new Formula(formula,typeName,this.sourceFile,this.lastLineNumber));
                this.annotatedItemName = null;
            }
        }
    }
}