package abc.ltl.formulaextraction;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.SAXException;

import polyglot.frontend.AbstractExtensionInfo;
import abc.ltl.Debug;
import abc.main.options.OptionsParser;
import de.tud.bat.classfile.structure.ClassFile;
import de.tud.bat.io.IO;
import de.tud.bat.io.xml.writer.XMLClassFileWriter;

/**
 * @author Eric Bodden
 *
 * Specification
 */
public class Specification {

    /**
     * Type name of the LTL annotation type.
     */
    public final static String LTL_TYPE_NAME = "rwth.i2.ltlrv.LTL";
    
    /**
     * Package name for formula classes to generate.
     */
    public final static String LTL_FORMULA_PACKAGE = "rwth.i2.ltlrv.formula";
    
    /**
     * Class name prefix for formula classes to generate.
     */
    public final static String LTL_FORMULA_CLASS_PREFIX = "Formula";

    /**
     * List of formulas that were extracted from the source bytecode.
     * For internal use only.
     */
    private static List<Formula> formulas = new ArrayList<Formula>();

	/**
	 * Counter for formula numbers. 
	 */
	private static int formulaId;
	
    /**
     * Extracts formulas from the given source files.
     * @param classIDs 
     * @param classPath
     * @throws IOException
     * @throws ClassNotFoundException
     */
    private static void init(List<String> classIDs, String[] classPath) throws IOException, ClassNotFoundException {
        final ByteArrayOutputStream bos = new ByteArrayOutputStream();
        final AnnotationExtractor annotationExtractor = new AnnotationExtractor();
        final ClassLoader classLoader = new URLClassLoader(pathToFileURLs(classPath));
        try {
            for(String fileName: classIDs) {
                classLoader.loadClass(fileName);
                ClassFile classFile = IO.readClassFile(classLoader, fileName);
                XMLClassFileWriter.serialize(classFile, bos);
                if(Debug.v().showExtractedXML) {
                	XMLClassFileWriter.serialize(classFile, System.err);
                }
                try {
                    SAXParser saxParser = SAXParserFactory.newInstance().newSAXParser();
                    ByteArrayInputStream bis = new ByteArrayInputStream(bos.toByteArray());
                    try {
                        saxParser.parse(bis,annotationExtractor);
                    } finally {
                        bis.close();
                    }
                } catch (ParserConfigurationException e) {
                    e.printStackTrace();
                } catch (SAXException e) {
                    e.printStackTrace();
                }
                bos.reset();
            }
        } finally {
            bos.close();
        }
    }
    
    /**
     * Converts file paths given as string into URLs for the URLClassLoader.
     * @throws MalformedURLException is a malformel URL would result.
     */
    private static URL[] pathToFileURLs(String[] paths) throws MalformedURLException {
        URL[] urls = new URL[paths.length];
  
        for (int i = 0; i < paths.length; i++) {
            String path = paths[i];
            File file = new File(path);
            path = file.getAbsolutePath();
            if(!file.exists()) {
                System.err.println("Warning: "+file.getAbsolutePath()+" does not exist.");
            }
            if(file.isDirectory()) {
                if(!(path.endsWith("/") || path.endsWith("\\"))) {
                    path += File.separator;
                }
            }
            urls[i] = new URL("file:///"+path);
        }
        
        
        return urls;
    }
    
   
    /*package */ static void addFormula(Formula formula) {
        formulas.add(formula);
    }

    public static String getFormulaId() {
        return LTL_FORMULA_CLASS_PREFIX + formulaId;		
    }

    private static boolean wasInitialized = false;
        
    public static void generateInitialSourceUnitsFromFormulas(AbstractExtensionInfo extensionInfo) {
    	assert wasInitialized==false; //should only be called once
    	assert wasInitialized = true;
    	
		//initialize/reset
		formulaId = 0;
		
    	@SuppressWarnings("unchecked")    	
        List<String> ltlPath = (List<String>) OptionsParser.v().ltlpath();
        List<String> classFiles =  classFilesInPath(ltlPath);
        
        try {
            Specification.init(classFiles, ltlPath.toArray(new String[0]));
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (ClassNotFoundException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
                
        List<Formula> formulas = Specification.getFormulas();
        
        for(Formula formula: formulas) {
			formulaId++;
            FormulaSource sourceUnit;
            try {
                byte[] content = classContentFromFormula(formula);
				//generate imaginary path to synthetic class
				String pathToClass = getFormulaId();
				if(LTL_FORMULA_PACKAGE.length()>0) {
					pathToClass = LTL_FORMULA_PACKAGE + "." + pathToClass;
				}
				pathToClass = pathToClass.replace('.',File.separatorChar);
				//generate source unit
                sourceUnit = new FormulaSource(content,pathToClass);
                if(Debug.v().showFormulaAspects) {
                    System.out.println(new String(content));
                }
                extensionInfo.addJob(sourceUnit);
            } catch (IOException e) {
                //cannot occur
                System.err.println("Internal error!!!");
                e.printStackTrace();
            }
        }
        
        //add shutdown hook
        String pathToClass =
        	"rwth.i2.ltlrv.management.ShutdownHook".replace('.',File.separatorChar);
        
        try {
			extensionInfo.addJob(
					new SyntheticSource(
							classContentForShutdownHookAspect(),
							pathToClass
					)
			);
        } catch (IOException e) {
            //cannot occur
            System.err.println("Internal error (adding shutdown hook)!!!");
            e.printStackTrace();
        }
        
    }
    
    private static byte[] classContentFromFormula(Formula formula) {
    	String aspectName = getFormulaId();
    	
        return (
        	//package decl
        	"package " + LTL_FORMULA_PACKAGE + ";\n"
            //aspect decl
        	+ "public privileged aspect " + aspectName + " {\n"
            + formula.getRepresentationForGrammar()
            + "\n}"
        ).getBytes();
    }
    
    /**
     * Returns code for the shutdown hook aspect.
     * This aspect notifies the runtime that the
     * application is going down.
     * The shutdown hook is added before the outermost execution of <code>main</code>.
     * @return code for the shutdown hook aspect
     */
    private static byte[] classContentForShutdownHookAspect() {
    	return ("package rwth.i2.ltlrv.management;  " + 
                "public aspect ShutdownHook {  " + 
                "" + 
                "   declare precedence: ShutdownHook,*;  " + 
                "" + 
                "   before(): staticinitialization(*) && !within(rwth.i2.ltlrv..*) {" + 
                "   }" + 
                "" + 
                "   public ShutdownHook(){  " + 
                "       Runtime.getRuntime().addShutdownHook(  " + 
                "           new Thread() {  " + 
                "               public void run() {  " + 
                "               VerificationRuntime.getInstance().tearDown();  " + 
                "               }" + 
                "           }" + 
                "       );  " +
                "   }  " + 
                "" + 
                "}" + 
                "").getBytes();
    }
    
    private static List<String> classFilesInPath(List<String> path) {
        List<String> result = new ArrayList<String>();
        for(String entry: path) {
            File possibleFolder = new File(entry);
            if(possibleFolder.exists() && possibleFolder.isDirectory()) {
                //accumulate class files
                findClassesInDir(possibleFolder.getAbsolutePath(), possibleFolder.getAbsolutePath(), result);
            }
        }
        return result;
    }
    
    private static void findClassesInDir(String dir, String baseDir, Collection<String> classFiles) throws IllegalArgumentException {
        if(baseDir.endsWith(File.separator)) {
            baseDir = baseDir.substring(0,baseDir.length()-1);
        }
        File file = new File(dir);
        File[] files = file.listFiles();
        if( files == null ) {
            return;
        }
        for( int i = 0; i < files.length; i++ ) {
            if(files[i].isDirectory()) findClassesInDir(files[i].getAbsolutePath(), baseDir, classFiles);
            else {
                String fileName = files[i].getAbsolutePath();
                if( fileName.endsWith(".class") ) {
                    fileName = fileName.substring(baseDir.length()+1);
                    //make class id from class file
                    String classId = fileName.replaceAll("\\\\",".");
                    classId = classId.replaceAll("/",".");
                    //replace dollar signs by dots
                    //classId = classId.replaceAll("\\$", ".");
                    //prune ".class" ending
                    classId = classId.substring(0,classId.length()-6);
                    classFiles.add(classId);
                }
            }
        }
    }
    
    /**
     * @return Returns the formulas.
     */
    public static List<Formula> getFormulas() {
        return formulas;
    }
}
