package abc.ltl.visit.transform;

import static abc.ltl.ast.EntryExit.Exit.ExitKind.RETURNING;
import static abc.ltl.ast.EntryExit.Exit.ExitKind.THROWING;
import static abc.ltl.visit.transform.RuntimeRepresentation.BINDINGS_MAP_ACTUAL_TYPE_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.BINDINGS_MAP_DECLARED_TYPE_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.BINDINGS_MAP_LOCAL_VARIABLE_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.BINDINGS_MAP_LOCAL_VARIABLE_PUT_METHOD;
import static abc.ltl.visit.transform.RuntimeRepresentation.CURRENT_PROPOSITIONS_ADD_METHOD;
import static abc.ltl.visit.transform.RuntimeRepresentation.CURRENT_PROPOSITIONS_CLEAR_METHOD;
import static abc.ltl.visit.transform.RuntimeRepresentation.CURRENT_PROPOSITIONS_LABEL;
import static abc.ltl.visit.transform.RuntimeRepresentation.DUMMY_VARIABLE_NAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.PROPOSITION_SPECIALIZE_METHOD;
import static abc.ltl.visit.transform.RuntimeRepresentation.RUNTIME_ENVIRONMENT_CLASSNAME;
import static abc.ltl.visit.transform.RuntimeRepresentation.RUNTIME_ENVIRONMENT_SINGLETON_ACCESS;
import static abc.ltl.visit.transform.RuntimeRepresentation.UPDATE_FORMULA;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import polyglot.ast.Block;
import polyglot.ast.CanonicalTypeNode;
import polyglot.ast.ClassDecl;
import polyglot.ast.Expr;
import polyglot.ast.Formal;
import polyglot.ast.Node;
import polyglot.ast.Stmt;
import polyglot.ast.TypeNode;
import polyglot.types.Flags;
import polyglot.types.PrimitiveType;
import polyglot.types.SemanticException;
import polyglot.types.TypeSystem;
import polyglot.util.Position;
import polyglot.util.TypedList;
import polyglot.visit.NodeVisitor;
import abc.aspectj.ast.AJNodeFactory;
import abc.aspectj.ast.AdviceDecl;
import abc.aspectj.ast.AdviceSpec;
import abc.aspectj.ast.AspectDecl;
import abc.aspectj.ast.ModifierPattern;
import abc.aspectj.ast.PCBinary;
import abc.aspectj.ast.Pointcut;
import abc.aspectj.ast.TypeFormalPattern;
import abc.ltl.ast.EntryExit;
import abc.ltl.ast.PointcutUtils;


public class AdviceGeneration extends PointcutsVisitor {

    /**
     * @param nf
     * @param ts
     */
    public AdviceGeneration(AJNodeFactory nf, TypeSystem ts) {
        super(nf,ts);
    }
    
    /**
     * Returns a position equal to <code>pos</code> but with a column incremented by 1.
     * @param pos
     * @return
     */
    private Position nextPos(Position pos) {
        return new Position(pos.file(),pos.line(),pos.column()+1);
    }    
    
    /**
     * {@inheritDoc}
     */
    @Override
    public Node leave(Node old, Node n, NodeVisitor v) {
        if(n instanceof AspectDecl) {
            ClassDecl aspect = (AspectDecl)n;

            //add initialization advice
            aspect = aspect.body(aspect.body().addMember(initializationAdvice(n.position())));
            
            boolean hasBeforeAdvice = false, hasAfterAdvice = false;
            
            //note that in the following we have to add "before"
            //advice first due to advice precedence rules
            
            //in order to have precedence being calculated correctly, we have to
            //simulate incremented position information
            Position pos = n.position();
            
            //generate "befor" advice to collect propositions
            for (Pointcut pc : pointcuts) {
                if(EntryExit.isEntry(pc)) {
                    //generate advice for this pointcut
                    AdviceDecl advice = propositionAdvice(pos,pc, EntryExit.getEntryExit(pc));
                    
                    //add advice to body
                    aspect = aspect.body(aspect.body().addMember(advice));
                    
                    //remember that we have a "before advice"
                    hasBeforeAdvice = true;
                    
                    //increment position to maintain correct lexical advice execution ordering
                    pos = nextPos(pos);
                }
            }
            
            //increment position to maintain correct lexical advice execution ordering
            pos = nextPos(pos);

            if(hasBeforeAdvice) {
                //generate "before" advice to trigger transitions
                AdviceDecl beforeAdvice = transitionAdvice(pos, true);
                aspect = aspect.body(aspect.body().addMember(beforeAdvice));
            }
            
            //generate "after" advice to collect propositions
            for (Pointcut pc : pointcuts) {
                if(EntryExit.isExit(pc)) {
                    //generate advice for this pointcut
                    AdviceDecl advice = propositionAdvice(pos,pc, EntryExit.getEntryExit(pc));
                    
                    //add advice to body
                    aspect = aspect.body(aspect.body().addMember(advice));

                    //remember that we have an "after advice"
                    hasAfterAdvice = true;

                    //increment position to maintain correct lexical advice execution ordering
                    pos = nextPos(pos);
                }
            }

            //increment position to maintain correct lexical advice execution ordering
            pos = nextPos(pos);

            if(hasAfterAdvice) {
                //generate "after" advice to trigger transitions
                AdviceDecl afterAdvice = transitionAdvice(pos, false);
                aspect = aspect.body(aspect.body().addMember(afterAdvice));
            }

            assert hasBeforeAdvice || hasAfterAdvice;
            
            n = aspect;
        }            
        return super.leave(old, n, v);
    }

    /**
     * @param pos TODO
     * @param pc
     * @param beforeAdvice 
     * @return
     * @throws SemanticException 
     */
    protected AdviceDecl propositionAdvice(Position pos, Pointcut pc, EntryExit.EntryExitKind entryExitInfo) {
        boolean isEntry = entryExitInfo instanceof EntryExit.Entry;        
        //retrieve boudn formals, first without those bound by if-pointcuts
        List<Formal> boundFormals = PointcutUtils.boundFormals(this.formula.getFormals(),pc,false);
        
        //advice header
        AdviceSpec adviceSpec;
        Formal returnOrThrowsFormal = null;
        
        if(!isEntry) {
            //if we have an "exit" event
            EntryExit.Exit exitInfo = (EntryExit.Exit) entryExitInfo; 
            
            //if we have "exit(...) returning x" or "exit(...) throwing x"
            if(exitInfo.getKind().equals(RETURNING)
                    || exitInfo.getKind().equals(THROWING)) {
                //determine name opf the variable or type
                TypeNode type = exitInfo.getType();
                String name = exitInfo.getTypeName();
                //see if this name is bound (then it is a variable, else a type) 
                TypeNode boundType = PointcutUtils.bindingNameToType(name, formula.getFormals().toArray(new Formal[0]));
                if(boundType==null) {
                    //the name represents a type, so create a dummy variable just to match on this type
                    name = DUMMY_VARIABLE_NAME;
                } else {
                    //the name represents a variable and boundType is the type of this variable
                    type = boundType;
                    //set formal for further processing
                    returnOrThrowsFormal = nf.Formal(pos,Flags.NONE,type,name);
                }
                //now "name" is the name of the variable to use and "type" holds its type
                
                //semantic check: check if the name bound by the returning/throwing construct is
                //not already bound by the other pointcut parameters
                if(PointcutUtils.bindingNameToType(name,boundFormals.toArray(new Formal[0]))!=null) {
                    //FIXME handle this error somehow
                }
                
                //generate advicespec
                if(exitInfo.getKind().equals(EntryExit.Exit.ExitKind.RETURNING)) {
                    //if "exit(...) returning x"
                    adviceSpec = nf.AfterReturning(
                            pos,
                            boundFormals,
                            nf.AdviceFormal(
                                    pos,
                                    Flags.NONE,
                                    type,
                                    name
                            ),
                            nf.CanonicalTypeNode(pos,ts.Void())
                    );
                } else {
                    //if "exit(...) throwing x"
                    adviceSpec = nf.AfterThrowing(
                            pos,
                            boundFormals,
                            nf.AdviceFormal(
                                    pos,
                                    Flags.NONE,
                                    type,
                                    name
                            ),
                            nf.CanonicalTypeNode(pos,ts.Void())
                    );
                }
                
            } else {
                //usual "after" advicespec
                adviceSpec = nf.After(pos,boundFormals,nf.CanonicalTypeNode(pos,ts.Void()));                
            }
        } else {
            //if "entry event" use usual "before" advicespec
            adviceSpec = nf.Before(pos,boundFormals,nf.CanonicalTypeNode(pos,ts.Void()));
        }
        
        //advice block
        Block block = nf.Block(pos);
        
        //initializer for map:
        //if the map has elements take a "real" map,
        //take the default empty map otherwise
        Expr bindingsDeclInit;
//        if((boundFormals.size()>0) || (returnOrThrowsFormal!=null)) {
            bindingsDeclInit = nf.New(
                    pos,
                    nf.AmbTypeNode(
                            pos,
                            BINDINGS_MAP_ACTUAL_TYPE_NAME
                    ),
                    Collections.EMPTY_LIST
            );
//        } else {
//            bindingsDeclInit = nf.Field(
//                    pos,
//                    nf.AmbReceiver(
//                            pos,
//                            "java.util.Collections"
//                    ),
//                    "EMPTY_MAP"
//            );
//        }
        
        //declare local variable for bindings map
        Stmt bindingsDecl = nf.LocalDecl(
                pos,
                Flags.FINAL,
                nf.AmbTypeNode(
                        pos,
                        BINDINGS_MAP_DECLARED_TYPE_NAME
                ),
                BINDINGS_MAP_LOCAL_VARIABLE_NAME,
                bindingsDeclInit
        );
        block = block.append(bindingsDecl);
        
        //code to add bindings to map
        for (Formal formal : boundFormals) {
            block = addBindingsCode(pos, block, formal);
        }
        //handle returning/throwing formals
        if(returnOrThrowsFormal!=null) {
            //if we have a declared return/exception formal
            //do the same for return/exception value
            block = addBindingsCode(pos, block, returnOrThrowsFormal);
        }
        
        //code to add proposition
        
        //get pointcut ID
        if(formula.propositionLabel(pc)==null) {
            throw new RuntimeException("Internal error: Pointcut has no ID! "+pc);
        }
        String propositionId = formula.propositionLabel(pc);
        //create method call
        Stmt addProp = nf.Eval(
                pos,
                nf.Call(
                        pos,
                        nf.AmbReceiver(
                                pos,
                                CURRENT_PROPOSITIONS_LABEL
                        ),
                        CURRENT_PROPOSITIONS_ADD_METHOD,
                        nf.Call(
                                pos,
                                nf.AmbReceiver(
                                        pos,
                                        propositionId                                           
                                ),
                                PROPOSITION_SPECIALIZE_METHOD,
                                nf.AmbExpr(
                                        pos,
                                        BINDINGS_MAP_LOCAL_VARIABLE_NAME
                                )
                        )
                )
        );
        block = block.append(addProp);
        
        //replace if(expr()) by if(true), since if-pointcuts are handled differently in our case
        pc = PointcutUtils.replaceIfPCsByIfTrue(pc,nf);
        
        //advice declaration
        return nf.AdviceDecl(
                pos,
                Flags.NONE,
                adviceSpec,
                Collections.EMPTY_LIST,
                pc,
                block
        );
    }

    /**
     * @param pos
     * @param block
     * @param formal
     * @return
     */
    protected Block addBindingsCode(Position pos, Block block, Formal formal) {
        final Expr exprToPut = autoBox(pos, formal);
    	
    	Stmt putBinding = nf.Eval(
                pos,
                nf.Call(
                        pos,
                        nf.AmbReceiver(
                                pos,
                                BINDINGS_MAP_LOCAL_VARIABLE_NAME
                        ),
                        BINDINGS_MAP_LOCAL_VARIABLE_PUT_METHOD,
                        nf.StringLit(
                                pos,
                                formal.name()
                        ),
                        exprToPut
                )
        );
        return block.append(putBinding);
    }

    /**
     * @param pos
     * @param formal
     * @return
     */
    private Expr autoBox(Position pos, Formal formal) {
        Expr unboxed = nf.AmbExpr(
                pos,
                formal.name()
        );
        
        if(formal.type() instanceof CanonicalTypeNode) {
            //if we have a canonical type, we have to autobox the value
            CanonicalTypeNode tn = (CanonicalTypeNode) formal.type();
            if(tn.type() instanceof PrimitiveType) {
                PrimitiveType pt = (PrimitiveType) tn.type();
                String boxTypeName;
                if(pt.kind().equals(PrimitiveType.BOOLEAN)) {
                    boxTypeName = "Boolean";
                } else if(pt.kind().equals(PrimitiveType.BYTE)) {
                    boxTypeName = "Byte";
                } else if(pt.kind().equals(PrimitiveType.CHAR)) {
                    boxTypeName = "Character";
                } else if(pt.kind().equals(PrimitiveType.DOUBLE)) {
                    boxTypeName = "Double";
                } else if(pt.kind().equals(PrimitiveType.FLOAT)) {
                    boxTypeName = "Float";
                } else if(pt.kind().equals(PrimitiveType.INT)) {
                    boxTypeName = "Integer";
                } else if(pt.kind().equals(PrimitiveType.LONG)) {
                    boxTypeName = "Long";
                } else if(pt.kind().equals(PrimitiveType.SHORT)) {
                    boxTypeName = "Short";
                } else {
                    throw new RuntimeException("Unexpected primitive type in 'Proposition advice' (autoboxing)");
                }
                TypeNode boxType = nf.AmbTypeNode(pos,"java.lang." + boxTypeName);
                List<Expr> args = new ArrayList<Expr>();
                args.add(unboxed);
                return nf.New(
                        pos,
                        boxType,
                        args
                );
            }
        } 
        return unboxed;
    }

    /**
     * @param pos 
     * @param pointcutsContainingTypes
     * @return
     */
    private Pointcut orPointcut(Position pos, List<Pointcut> disjuncts) {
        assert disjuncts.size()>0;
        if(disjuncts.size()==1) {
            return disjuncts.get(0);            
        } else {
            Pointcut p1 = disjuncts.remove(0);
            Pointcut p2 = disjuncts.remove(0);
            disjuncts.add(0,nf.PCBinary(pos, p1, PCBinary.COND_OR, p2));
            return orPointcut(pos, disjuncts);
        }
    }

    /**
     * @param pos
     * @param beforeAdvice
     * @return
     */
    protected AdviceDecl transitionAdvice(Position pos, boolean beforeAdvice) {
        Collection<Pointcut> beforeAfterPointcuts = new ArrayList<Pointcut>();

        //filter for correct pointcuts
        for (Pointcut pointcut : this.pointcuts) {
            if((beforeAdvice && EntryExit.isEntry(pointcut))
            ||(!beforeAdvice && EntryExit.isExit(pointcut))) {
                beforeAfterPointcuts.add(pointcut);
            }
        }
        
        List<Pointcut> pointcutsContainingTypes = new ArrayList<Pointcut>();

        //replace all bindings inpointcuts by their types
        for (Pointcut pointcut : beforeAfterPointcuts) {
            Pointcut newPc = PointcutUtils.replaceBindingsByTypes(this.formula.getFormals(), pointcut);
            pointcutsContainingTypes.add(newPc);
        }
        
        //build an "or"-pointcut over those
        Pointcut combinedPC = orPointcut(pos, pointcutsContainingTypes);
        
        //replace if(expr()) by if(true), since if-pointcuts are handled differently in our case
        combinedPC = PointcutUtils.replaceIfPCsByIfTrue(combinedPC, nf);

        //build the block        
        Block block = nf.Block(pos);
        
        //update the formula
        block = block.append(
                nf.Eval(
                        pos,
                        nf.Call(
                                pos,
                                nf.Call(
                                        pos,
                                        nf.AmbReceiver(
                                                pos,
                                                RUNTIME_ENVIRONMENT_CLASSNAME
                                        ),
                                        RUNTIME_ENVIRONMENT_SINGLETON_ACCESS
                                ),                                    
                                UPDATE_FORMULA,
                                nf.StringLit(
                                        pos,
                                        enclosingAspect.name()
                                ),
                                nf.AmbExpr(
                                        pos,
                                        CURRENT_PROPOSITIONS_LABEL
                                )
                        )
                )
        );     
    
        //clear the proposition set
        block = block.append(
                nf.Eval(
                        pos,
                        nf.Call(
                                pos,
                                nf.AmbReceiver(
                                        pos,
                                        CURRENT_PROPOSITIONS_LABEL
                                ),
                                CURRENT_PROPOSITIONS_CLEAR_METHOD
                        )
                )
        );        

        //build the advice spec
        AdviceSpec adviceSpec;
        if(beforeAdvice) {
            adviceSpec = nf.Before(pos,Collections.EMPTY_LIST,nf.CanonicalTypeNode(pos,ts.Void()));
        } else {
            adviceSpec = nf.After(pos,Collections.EMPTY_LIST,nf.CanonicalTypeNode(pos,ts.Void()));
        }        
        
        //return the full advice declaration
        return nf.AdviceDecl(
                pos,
                Flags.NONE,
                adviceSpec,
                Collections.EMPTY_LIST,
                combinedPC,
                block
        );
    }
    
    /**
     * Generates an advice with emtpy block which is to be executed before any main method.
     * This is to just have the aspect class loaded at runtime
     * (actually onle necessary for existential formulae (F,U), since those are nonfinal).
     * @param pos
     * @return
     */
    @SuppressWarnings("unchecked")
    protected AdviceDecl initializationAdvice(Position pos) {
    	//list of modifiers: public static
    	List modifiers = new ArrayList();;
    	modifiers.add(
    			nf.ModifierPattern(
    					pos,
    					Flags.PUBLIC.set(Flags.STATIC),
    					true
    			)
    	);
		modifiers = new TypedList(modifiers,ModifierPattern.class,true);
    	
    	//list of arguments: String[]
		List args = new ArrayList();
		args.add(
				nf.TypeFormalPattern(
						pos,
						nf.TPEArray(
								pos,
								nf.TPERefTypePat(
										pos,
										nf.RTPName(
												pos,
												nf.SimpleNamePattern(pos,"String")
										)
								),
								1
						)
				)
		);
		//seal list
		args = new TypedList(args,TypeFormalPattern.class,true);
		
		return nf.AdviceDecl(
    			pos,
    			Flags.NONE,
    			nf.Before(
    					pos,
    					Collections.EMPTY_LIST,
    					nf.CanonicalTypeNode(pos,ts.Void())
    			),
    			Collections.EMPTY_LIST,
    			nf.PCExecution(
    					pos,
    					nf.MethodPattern(
    							pos,
    							modifiers, /* public static */
    							nf.TPEType(pos,nf.CanonicalTypeNode(pos,ts.Void())), /* void */
    							nf.ClassTypeDotId( /* *.main */
    									pos,
    									nf.CPEUniversal(pos),
    									nf.SimpleNamePattern(pos,"main")
    							),
    							args,
    							Collections.EMPTY_LIST
    					)
    			),
    			nf.Block(pos) //empty block
    	);    	
    }

    
}
