package abc.ltl.visit.transform;

import static abc.ltl.visit.transform.RuntimeRepresentation.CURRENT_PROPOSITIONS_LABEL;
import static abc.ltl.visit.transform.RuntimeRepresentation.PROPOSITION_SET_CLASS_NAME;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import abc.ltl.ast.PCFormula;

import polyglot.ast.AmbExpr;
import polyglot.ast.AmbReceiver;
import polyglot.ast.Call;
import polyglot.ast.ClassMember;
import polyglot.ast.Expr;
import polyglot.ast.FieldDecl;
import polyglot.ast.Node;
import polyglot.ast.NodeFactory;
import polyglot.ast.TypeNode;
import polyglot.types.Flags;
import polyglot.util.Position;
import polyglot.visit.NodeVisitor;

public class MultiThreadingSecure extends NodeVisitor {

    private NodeFactory nf;
    
    private PCFormula enclosingFormula;
    
    /**
     * @param nf
     */
    public MultiThreadingSecure(NodeFactory nf) {
        super();
        this.nf = nf;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Node leave(Node old, Node n, NodeVisitor v) {
        if(n instanceof PCFormula) {
            
            enclosingFormula = (PCFormula)n;
            
        } else if(
                (enclosingFormula !=null)
                && enclosingFormula.isPrepareForMultiThreading()
                && (n instanceof FieldDecl)) {
            
            FieldDecl fd = (FieldDecl)n;
            
            //if we have the field declaration for the set of current propositions
            if(fd.name().equals(CURRENT_PROPOSITIONS_LABEL)) {
                
                final Position pos = fd.position();
                
                //type of the ThreadLocal
                TypeNode threadLocal = nf.AmbTypeNode(
                        pos,
                        "ThreadLocal"
                );
                
                //class members for the new ThreadLocal
                List<ClassMember> threadLocalClassMembers = new ArrayList<ClassMember>();
                threadLocalClassMembers.add(
                        nf.MethodDecl(
                                pos,
                                Flags.PROTECTED,
                                nf.AmbTypeNode(
                                        pos,
                                        "Object"
                                ),
                                "initialValue",
                                Collections.EMPTY_LIST,
                                Collections.EMPTY_LIST,
                                nf.Block(
                                        pos,
                                        nf.Return(
                                                pos,
                                                nf.New(
                                                        pos,
                                                        nf.AmbTypeNode(
                                                                pos,
                                                                PROPOSITION_SET_CLASS_NAME
                                                        ),
                                                        Collections.EMPTY_LIST
                                                )
                                        )
                                )
                        )
                );                
                
                //initializer for ThreadLocal
                Expr init = nf.New(
                        pos,
                        threadLocal,
                        Collections.EMPTY_LIST,
                        nf.ClassBody(
                                pos,
                                threadLocalClassMembers                       
                        )
                );
                
                //new field declaration
                n = nf.FieldDecl(
                        pos,
                        fd.flags(),
                        threadLocal,
                        fd.name(),
                        init
                );
            }
            
        } else if(
                (enclosingFormula !=null)
                && enclosingFormula.isPrepareForMultiThreading()
                && (n instanceof Call)) {

            //replace calls to the set of current propositions to calls to the treadlocal
            Call origCall = (Call) n;
            
            final Position pos = origCall.position();

            //if we have a call to the set of current propositions
            if(origCall.target() instanceof AmbReceiver) {
                if(((AmbReceiver)origCall.target()).name().equals(CURRENT_PROPOSITIONS_LABEL)) {
                    //substitute this call by a call to the ThreadLocal instead 
                    
                    n = origCall.target(
                            nf.Cast(
                                    pos,
                                    nf.AmbTypeNode(
                                            pos,
                                            PROPOSITION_SET_CLASS_NAME
                                    ),
                                    nf.Call(
                                            pos,
                                            origCall.target(),
                                            "get"
                                    )                            
                            )
                    );
                }
            }
        } else if(
                (enclosingFormula !=null)
                && enclosingFormula.isPrepareForMultiThreading()
                && (n instanceof AmbExpr)) {

            //replace the ambiguous expressions in the same way 
            AmbExpr origExpr = (AmbExpr) n;
            
            final Position pos = origExpr.position();

            if(origExpr.name().equals(CURRENT_PROPOSITIONS_LABEL)) {
                
                n = nf.Cast(
                        pos,
                        nf.AmbTypeNode(
                                pos,
                                PROPOSITION_SET_CLASS_NAME
                        ),
                        nf.Call(
                                pos,
                                origExpr,
                                "get"
                        )                            
                );
            }
        }            
        return super.leave(old, n, v);
    }
}
